/* クリップボード2
 *
 * 左クリックで、CLIPBOARD から文字列を取得して表示。
 *
 * 閉じるボタンで終了 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <poll.h>
#include <errno.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include "util.h"

enum
{
	_ATOM_CLIPBOARD,
	_ATOM_MY_SELECTION,
	_ATOM_TARGETS,
	_ATOM_UTF8_STRING,
	_ATOM_COMPOUND_TEXT,
	_ATOM_TEXT,

	_ATOM_NUM
};

const char *g_names[] = {
	"CLIPBOARD", "_MY_SELECTION_", "TARGETS", "UTF8_STRING",
	"COMPOUND_TEXT", "TEXT"
};

Atom atoms[_ATOM_NUM];

/* t1 - t2 のミリ秒を取得 */

static long _get_diff_time(struct timespec *t1,struct timespec *t2)
{
	long sec,ns;

	sec = t1->tv_sec - t2->tv_sec;
	
	if(t1->tv_nsec >= t2->tv_nsec)
		ns = t1->tv_nsec - t2->tv_nsec;
	else
	{
		sec--;
		ns = 1000 * 1000 * 1000 - t2->tv_nsec + t1->tv_nsec;
	}

	return sec * 1000 + ns / (1000 * 100);
}

/* SelectionNotify イベントを受信 */

static int _recv_selection_notify(Display *disp,XEvent *ev)
{
	struct pollfd fds;
	struct timespec ts,tsend;
	int ret;
	long ms;

	fds.fd = ConnectionNumber(disp);
	fds.events = POLLIN;

	//タイムアウト終了時間 (+1s)

	clock_gettime(CLOCK_MONOTONIC, &tsend);

	tsend.tv_sec += 1;

	//

	while(1)
	{
		//キュー内にあるか (削除 & ブロックなし)

		if(XCheckTypedEvent(disp, SelectionNotify, ev))
			break;

		//タイムアウトまでの時間

		clock_gettime(CLOCK_MONOTONIC, &ts);

		ms = _get_diff_time(&tsend, &ts);
		if(ms < 0) return 1;

		//イベントを受信するまで待つ

		ret = poll(&fds, 1, ms);

		if(ret == 0)
			return 1; //タイムアウト
		else if(ret == -1)
			return 1; //エラー
	}

	return 0;
}

/* 指定ターゲットでテキスト読み込み */

static int _read_text(Display *disp,Window win,Atom target,const char *target_name)
{
	XEvent ev;
	XTextProperty tp;
	Atom type;
	char *buf,**list,*name;
	int size,num;

	printf("--- [%s] ---\n", target_name);

	XConvertSelection(disp, atoms[_ATOM_CLIPBOARD],
		target, atoms[_ATOM_MY_SELECTION], win, CurrentTime);

	if(_recv_selection_notify(disp, &ev))
	{
		printf("<no SelectionNotify>\n");
		return 1;
	}

	if(ev.xselection.property == None)
	{
		printf("<failed>\n");
		return 1;
	}

	//プロパティ読み込み

	buf = (char *)read_prop8(win, atoms[_ATOM_MY_SELECTION], &type, &size);

	XDeleteProperty(disp, win, atoms[_ATOM_MY_SELECTION]);
	
	if(!buf)
	{
		printf("<no property>\n");
		return 1;
	}

	//各タイプごと

	if(type == atoms[_ATOM_COMPOUND_TEXT])
	{
		//COMPOUND_TEXT

		tp.value = (unsigned char *)buf;
		tp.encoding = type;
		tp.format = 8;
		tp.nitems = size;

		if(XmbTextPropertyToTextList(disp, &tp, &list, &num) == Success)
		{
			printf("<COMPOUND_TEXT>\n%s\n", *list);
			XFreeStringList(list);
		}
	}
	else if(type == XA_STRING)
	{
		printf("<STRING>\n%s\n", buf);
	}
	else if(type == atoms[_ATOM_UTF8_STRING])
	{
		printf("<UTF8_STRING>\n%s\n", buf);
	}
	else
	{
		name = XGetAtomName(disp, type);
		if(name)
		{
			printf("<%s>\n%s\n", name, buf);
			XFree(name);
		}
	}

	free(buf);

	return 0;
}

int main(int argc,char **argv)
{
	Display *disp;
	Window win;
	XEvent ev;
	
	disp = XOpenDisplay(NULL);
	if(!disp) return 1;

	set_display(disp);
	init_locale();

	XInternAtoms(disp, (char **)g_names, _ATOM_NUM, False, atoms);

	win = create_test_window2(disp, 200, 200, 0, ButtonPress);

	//イベント

	XMapWindow(disp, win);

	while(1)
	{
		XNextEvent(disp, &ev);

		if(event_quit(&ev)) break;

		switch(ev.type)
		{
			case ButtonPress:
				if(ev.xbutton.button == Button1)
				{
					_read_text(disp, win, atoms[_ATOM_UTF8_STRING], "UTF8_STRING");
					_read_text(disp, win, atoms[_ATOM_COMPOUND_TEXT], "COMPOUND_TEXT");
					_read_text(disp, win, XA_STRING, "STRING");
					_read_text(disp, win, atoms[_ATOM_TEXT], "TEXT");
				}
				break;
		}
	}

	XCloseDisplay(disp);

	return 0;
}
